#pragma once

#include <qlat-utils/show.h>

namespace qlat
{  //

const bool IS_USING_INTERPOLATION = true;

const bool IS_USING_AGGRESSIVE_SUBTRACTION = true;

const bool IS_USING_MORE_AGGRESSIVE_SUBTRACTION = true;

const bool IS_USING_MUON_LINE_INTERPOLATION = true;

const Int IS_USING_ONLY_SUBTRACTION = 0;

const RealD DISTANCE_LIMIT = 6.0;

}  // namespace qlat
